#!/usr/bin/env python3
#
# Author: Yipeng Sun <syp at umd dot edu>
# License: BSD 2-clause
# Last Change: Fri May 01, 2020 at 05:06 PM +0800

__name__ = 'pyBabyMaker'
__version__ = '0.2.4'  # This version is fixed.
